/*
 ## Cypress FX3 Firmware Example Source (cyfx3_uvc.c)
 ## ===========================
 ##
 ##  Copyright Cypress Semiconductor Corporation, 2013-2014,
 ##  All Rights Reserved
 ##  UNPUBLISHED, LICENSED SOFTWARE.
 ##
 ##  CONFIDENTIAL AND PROPRIETARY INFORMATION
 ##  WHICH IS THE PROPERTY OF CYPRESS.
 ##
 ##  Use of this file is governed
 ##  by the license agreement included in the file
 ##
 ##     <install>/license/license.txt
 ##
 ##  where <install> is the Cypress software
 ##  installation root directory path.
 ##
 ## ===========================
 *  Created on: 30-June-2015
 *  Author: mddd
 */

/* This application example implements a USB UVC 1.5 compliant video camera on the FX3.
 */

#include "cyu3system.h"
#include "cyu3os.h"
#include "cyu3dma.h"
#include "cyu3error.h"
#include "cyu3usb.h"
#include "cyu3i2c.h"
#include "cyu3spi.h"
#include "cyu3uart.h"
#include "cyu3gpio.h"
#include "cyu3utils.h"
#include "cyu3pib.h"
#include "cyu3socket.h"
#include "sock_regs.h"
#include "cyfx3_uvc.h"
#include "cyu3mipicsi.h"
#include "cyfx3_videocontrol.h"
#include "cyfx3_videostreaming.h"
#include "cyu3_isp.h"
#include "cyfxgpif2config.h"
#include "stddef.h"
#include "string.h"

#define CY_FX_COMMIT_BUFFER_FAIL_THRESHOLD (50)
#define START_PAGE_ADDR		(1280)
#define CY_U3P_S0_CFG		(1 << 4)
#define CY_U3P_S1_CFG_POS	(8)


CyU3PEvent glFx3Event; 			 /* Application Event Group */
static CyU3PThread uvcAppThread; /* Application thread used for streaming from the GPIF interface to USB */

static volatile uint8_t  glActiveSocket = 0;
static volatile uint32_t glDMATxCount = 0; 	 /* Counter used to count the Dma Transfers */
static volatile uint32_t glDmaDone = 0;

static volatile CyBool_t glLpmDisable = CyTrue;		/* Flag used to Enable/Disable low USB 3.0 LPM */
static volatile CyBool_t glFrameEnd   = CyFalse;

static CyBool_t glHitFV = CyFalse;                  /* Flag used for state of FV signal. */
static CyBool_t glIsClearFeature = CyFalse;         /* Flag to signal when AppStop is called from the ClearFeature request. Needed to clear Endpoint data toggles. */
static CyBool_t glRstDmaOnEPClearFeature = CyTrue;  /* Flag to stop streaming data after a clear feature is received on the BULK IN EP.Note that this flag is also used to handle suspend mode event */
static CyBool_t glReceivedSuspEvent   = CyFalse;    /*Suspend event flag*/
static CyBool_t glIsConfigured = CyFalse; 			/* Whether Application is in configured state or not */
static CyBool_t glIsStreamingStarted = CyFalse; 	/* Whether streaming has started - Used for MAC OS support*/

CyBool_t glIsApplnActive = CyFalse;                 /* Whether the GPIF->USB application is active or not. */

uint8_t  glFx3FwVer[5] = {0}; /* Variable to store FX3 FW version */
uint8_t  toggle = 1;
uint8_t  glDmaResetFlag = 0; /*Variable to track whether the reason for DMA Reset is Frame Timer overflow or Commit Buffer Failure*/
uint8_t  glCurrentVideoFormatIndex = 1;				/*Variables to monitor current format Index*/
uint8_t  glCurrentVideoFrameIndex  = 1;				/*Variables to monitor current frame Index*/
uint16_t glFrameTimerPeriod = 200;					/* Maximum frame transfer time in milli-seconds. */
uint16_t glSpiPageSize = 0x100;
uint32_t glCurrentFrameInterval = 0x51615;			/*Variables to monitor current frame Interval*/
uint32_t glFrameSizeExpected, glFrameSizeTransferred, glFrameSize; /* Variable to monitor the Frame Size of the YUV resolution being transferred. */
uint32_t previous_size = 0, current_size = 0;		/*Variables used to track the frame size using buffer size */


static uint8_t glCommitCtrl[FX3_UVC_MAX_PROBE_SETTING_ALIGNED]; /* Video Probe Commit Control */
static CyU3PDmaMultiChannel glChHandleUVCStream; 	/* DMA Channel Handle for UVC Stream */
static CyU3PDmaChannel glChHandleStatusIntrFC;      /* DMA channel handle for FC's status INT EP */


/* Mac OS does not send EP Clear feature when the app is closed. It just stops issuing IN tokens. So buffer commit failures
 * can be counted and if it reaches beyond a limit, streaming can be stopped. Buffer commit failure code can be cleared
 * on DMA Consumer event so that the limit is not reached under streaming conditions.  */
uint8_t glCommitBufferFailureCount = 0;

/* Variables to keep track of type-C enumeration mode */
#ifdef TYPE_C_ENABLED
    CyBool_t U3_EVENT = CyFalse;
    CyBool_t U3_TRUE = CyTrue;
#endif

#ifdef STILL_CAPTURE_ENABLE
    static CyU3PEvent glStillImageEvent; 				/*Still image event group. */
    static uint8_t glStillCommitCtrl[FX3_UVC_MAX_PROBE_SETTING_ALIGNED]; /*Still Commit Control Array*/
    static uint8_t glStillReqBuff[2] ={0}; 						/*Still Trigger Control Array*/
    static uint8_t glCurrentStillFormatIndex = 1, glCurrentStillFrameIndex = 1; /*Frame Index for Still Capture*/
    volatile CyBool_t glStillFlag = CyFalse; 					/*Still Image Event Flag*/
    volatile CyBool_t glStillHeader = 0;
    #define FX3_UVC_MAX_STILL_PROBE_SETTING         (11)               /* Maximum number of bytes in Still Probe Control */
    #define FX3_UVC_HEADER_STI                      (uint8_t)(1 << 5)  /* Still Image Indication in the UVC Header*/
    #define FX3_STILL_IMAGE_EVENT                   (1 << 7)           /*Event used to Still Image Capture*/

    /* Still Probe Control Setting */
    uint8_t glStillProbeCtrl[FX3_UVC_MAX_STILL_PROBE_SETTING] =
    {
            0x01, 										/* Use 1st Video format index */
            0x01, 										/* Use 1st Video frame index */
            0x00, 										/* Compression quality */
            0x00, 0xD8, 0xF3, 0x00, 					/* Max video frame size in bytes (Highest resolution frame size) */
    #ifdef USE_ISOCHRONOUS_ENDPOINT
            0x00, 0xB4, 0x00, 0x00 						/* No. of bytes device can rx in single payload = 56KB */
    #else
            0x00, 0xE0, 0x00, 0x00 						/* No. of bytes device can rx in single payload = 56KB */
    #endif
    };
#endif

#ifdef FRAME_TIMER_ENABLE
static CyU3PTimer FrameTimer;							/* Timer used to track frame transfer time. */

static void CyFx3UvcAppProgressTimer(uint32_t arg)
{
    /* Below check can be Suspend receive flag */
    if (glRstDmaOnEPClearFeature) /* De-initialize the ISP and enter low power mode */
    {
        CyU3PEventSet(&glFx3Event, FX3_USB_SUSP_EVENT_FLAG, CYU3P_EVENT_OR);
    }
    else /* This frame has taken too long to complete. Notify the thread to abort the frame and restart streaming. */
    {
        glDmaResetFlag = 1;

        CyU3PEventSet(&glFx3Event, FX3_DMA_RESET_EVENT, CYU3P_EVENT_OR);
        CyU3PDebugPrint(4, "\n\rPeriod = %d", glFrameTimerPeriod);
    }
}

#endif


#ifdef PRINT_FRAME_INFO
/*Variables to debug print the frame information like Producer Count, Consumer Count, Partial Packet size (in Bytes), Frame size (in Bytes) and FPS*/
uint32_t glProd = 0;
uint32_t glCons = 0;
uint32_t glProdCount = 0;
uint32_t glConsCount = 0;
volatile CyBool_t glPrintFlag = 0;
volatile uint32_t glFrameCount = 0;
volatile uint32_t glPartialBufSize = 0;
volatile uint32_t glTimeStart = 0;
volatile uint32_t glTimeEnd = 0;
volatile uint32_t glPeriod = 0;
volatile CyBool_t glFpsFlag = 0;
volatile CyBool_t glGetTimeFlag = 0;
#endif



/*Structure to store FX3 firmware version*/
void CyFx3GetFirmwareVersion()
{
    glFx3FwVer[0] = FX3_FW_VERSION_MAJOR;
    glFx3FwVer[1] = FX3_FW_VERSION_MINOR;
    glFx3FwVer[2] = FX3_FW_BUILD_MONTH;
    glFx3FwVer[3] = FX3_FW_BUILD_DAY;
    glFx3FwVer[4] = FX3_FW_BUILD_YEAR;
}

/*FX3 SPI Block Initialization. This is required for ISP Initialization to go through, since ISP in initialized through SPI Interface*/
CyU3PReturnStatus_t CyFx3SpiInit(CyBool_t cpol, CyBool_t cpha)
{
    CyU3PSpiConfig_t spiConfig;
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    /* Start the SPI module and configure the master. */
    status = CyU3PSpiInit();
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "Spi Init Failed %d\n\r", status);
        return status;
    }

    /* Start the SPI master block. Run the SPI clock at 8MHz
     * and configure the word length to 8 bits. Also configure
     * the slave select using FW. */
    CyU3PMemSet((uint8_t *)&spiConfig, 0, sizeof(spiConfig));
    spiConfig.isLsbFirst = CyFalse;
    spiConfig.cpol = cpol;
    spiConfig.ssnPol = CyFalse;
    spiConfig.cpha = cpha;
    spiConfig.leadTime = CY_U3P_SPI_SSN_LAG_LEAD_HALF_CLK;
    spiConfig.lagTime = CY_U3P_SPI_SSN_LAG_LEAD_HALF_CLK;
    spiConfig.ssnCtrl = CY_U3P_SPI_SSN_CTRL_FW;
    spiConfig.clock = 30000000;
    spiConfig.wordLen = 8;

    status = CyU3PSpiSetConfig(&spiConfig, NULL);

    return status;
}

/* Application critical error handler */
void CyFx3UvcAppErrorHandler(CyU3PReturnStatus_t status /* API return status */
)
{
    /* Application failed with the error code status */

    /* Add custom debug or recovery actions here */
    CyU3PDebugPrint(4,"\r\nEntering Error Handler, code will get stuck here..status = 0x%x",status);

    /* Loop indefinitely */
    for (;;)
    {
        /* Thread sleep : 100 ms */
        CyU3PThreadSleep(100);
    }
}

void CyFx3ApplnDmaCallback(CyU3PDmaMultiChannel *chHandle,CyU3PDmaCbType_t type, CyU3PDmaCBInput_t *input)
{
    CyU3PReturnStatus_t status;
    CyU3PDmaBuffer_t dmaBuffer;

    if (type == CY_U3P_DMA_CB_PROD_EVENT)
    {
        CyU3PTimerModify(&FrameTimer,glFrameTimerPeriod,0);
        CyU3PTimerStart(&FrameTimer);
        status = CyU3PDmaMultiChannelGetBuffer(chHandle, &dmaBuffer, CYU3P_NO_WAIT);
        if (status != CY_U3P_SUCCESS)
        {
            glDmaResetFlag = 2;
            CyU3PEventSet(&glFx3Event, FX3_DMA_RESET_EVENT, CYU3P_EVENT_OR);
            return;
        }
        if (glStillHeader==1)
        {
            dmaBuffer.buffer[1] ^=FX3_UVC_HEADER_STI; /* Modify UVC Header to notify still capture frame*/
        }
        status = CyU3PDmaMultiChannelCommitBuffer(chHandle, dmaBuffer.count, 0);
        if (status != CY_U3P_SUCCESS)
        {
            CyU3PEventSet(&glFx3Event, FX3_DMA_RESET_EVENT, CYU3P_EVENT_OR);
            return;
        }
#ifdef PRINT_FRAME_INFO
        /*1st Packet of frame is received. Get the Start time to calculate the FPS.*/
        if (glGetTimeFlag == 0)
        {
            glTimeStart = CyU3PGetTime();
            glGetTimeFlag = 1;
        }
        glProd++;
#endif
        current_size = input->buffer_p.count - FX3_UVC_HEADER_LENGTH;
        glFrameSizeTransferred += (input->buffer_p.count - FX3_UVC_HEADER_LENGTH);

        if(current_size < previous_size)
        {
            glFrameSize = glFrameSizeTransferred;
            glFrameSizeTransferred = 0;
#ifdef STILL_CAPTURE_ENABLE
            if(glStillFlag)
            {
                CyU3PEventSet(&glStillImageEvent, FX3_STILL_IMAGE_EVENT, CYU3P_EVENT_OR);
            }
#endif

#ifdef PRINT_FRAME_INFO
            /*Collect the frame information when frame ends*/
            glFrameCount++;
            glPartialBufSize = (input->buffer_p.count - FX3_UVC_HEADER_LENGTH);
            glConsCount = glCons;
            glProdCount = glProd;
            glPrintFlag = 1;

            /*Frame has ended. Clear the producer and consumer counts*/
            glProd = 0;
            glCons = 0;
            glHitFV = CyFalse;
            glDMATxCount = 0;

#endif
#ifdef FRAME_TIMER_ENABLE
            CyU3PTimerStop(&FrameTimer);
#endif
#if FPS
        if ((glFrameCount % 60) == 0 && glFrameCount != 0)
        {
            glTimeEnd = CyU3PGetTime();
            glPeriod = glTimeEnd - glTimeStart;
            glFpsFlag = 1;
            glGetTimeFlag = 0;
        }
#endif
        }
        previous_size = current_size;
    }
    else if (type == CY_U3P_DMA_CB_CONS_EVENT)
    {
#ifdef PRINT_FRAME_INFO
        glCons++;
#endif
    }
}


/* This function starts the video streaming application. It is called
 * when there is a SET_INTERFACE event for alternate interface 1
 * (in case of UVC over Isochronous Endpoint usage) or when a
 * COMMIT_CONTROL(SET_CUR) request is received (when using BULK only UVC).
 */
CyU3PReturnStatus_t
CyFx3UvcAppStart(uint8_t EPNumber)
{
#ifdef FX3_DEBUG_ENABLE_LEVEL2
    uint8_t SMState = 0;
#endif
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    CyU3PDmaMultiChannelConfig_t dmaCfg;

    glIsApplnActive = CyTrue;

    glDmaDone = 0;
    glDMATxCount = 0;
    glHitFV = CyFalse;
    glLpmDisable = CyTrue;
#ifdef PRINT_FRAME_INFO
    /*Variables to debug print the frame information like Producer Count, Consumer Count, Partial Packet size (in Bytes), Frame size (in Bytes) and FPS*/
    glProd = 0;
    glCons = 0;

    glPrintFlag = 0;
    glFrameCount = 0;
    glPartialBufSize = 0;
    glTimeStart = 0;
    glTimeEnd = 0;
    glPeriod = 0;
    glFpsFlag = 0;
    glGetTimeFlag = 0;
#endif
    glFrameSizeTransferred = 0;

#ifdef FRAME_TIMER_ENABLE
    CyU3PTimerStop(&FrameTimer);
    CyU3PTimerModify(&FrameTimer,glFrameTimerPeriod,0);
    CyU3PTimerStart(&FrameTimer);
#endif

    /* Place the EP in NAK mode before cleaning up the pipe. */
    CyU3PUsbSetEpNak(FX3_EP_BULK_VIDEO_1, CyTrue);
    CyU3PBusyWait(FX3_EP_BUSY_WAIT);

    /* Reset USB EP and DMA */
    CyU3PUsbFlushEp(FX3_EP_BULK_VIDEO_1);
    status = CyU3PDmaMultiChannelReset(&glChHandleUVCStream);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStart:ChannelReset Err = 0x%x", status);
        return status;
    }


    status = CyU3PDmaMultiChannelDestroy(&glChHandleUVCStream);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStart:ChannelReset Err = 0x%x", status);
        return status;
    }

    /* Configure MANUAL MANY TO ONE DMA channel for GPIF to USB transfer */
    CyU3PMemSet ((uint8_t *)&dmaCfg, 0, sizeof (dmaCfg));
    dmaCfg.size = 42*1024;
    dmaCfg.count = 2;
    dmaCfg.validSckCount = 2;
    dmaCfg.prodSckId[0] = CY_U3P_PIB_SOCKET_0;
    dmaCfg.prodSckId[1] = CY_U3P_PIB_SOCKET_1;
    dmaCfg.consSckId[0] = FX3_EP_VIDEO_CONS_SOCKET_1;
    dmaCfg.dmaMode = CY_U3P_DMA_MODE_BYTE;
    dmaCfg.notification = CY_U3P_DMA_CB_PROD_EVENT | CY_U3P_DMA_CB_CONS_EVENT;
    dmaCfg.cb = CyFx3ApplnDmaCallback;
    dmaCfg.prodHeader = 0;
    dmaCfg.prodFooter = 0;
    dmaCfg.consHeader = 0;
    dmaCfg.prodAvailCount = 0;


    status = CyU3PDmaMultiChannelCreate(&glChHandleUVCStream, CY_U3P_DMA_TYPE_MANUAL_MANY_TO_ONE, &dmaCfg);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStart:DmaMultiChannelCreate Err = 0x%x",
                status);
        CyU3PDeviceReset(CyFalse);
    }

    CyU3PThreadSleep(10);

    status = CyU3PDmaMultiChannelSetXfer(&glChHandleUVCStream, 0, 0);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStart:SetXfer Err = 0x%x", status);
        return status;
    }


    CyU3PUsbSetEpNak(FX3_EP_BULK_VIDEO_1, CyFalse);
    CyU3PBusyWait(FX3_EP_BUSY_WAIT);

#ifdef FX3_DEBUG_ENABLE_LEVEL2
    CyU3PGpifGetSMState(&SMState);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStart:SMState = 0x%x", SMState);
#endif


    glFrameEnd = CyFalse;
    glActiveSocket = 0;

    CyU3PUsbLPMDisable();
    CyU3PUsbSetLinkPowerState(CyU3PUsbLPM_U0);
    CyU3PThreadSleep(1);

    CyU3PGpioSetValue(SLOE_GPIO, CyFalse);
    CyU3PGpioSetValue(SLOE_GPIO, CyTrue);

    status = CyU3PGpifSMSwitch(FX3_INVALID_GPIF_STATE, 0,
            FX3_INVALID_GPIF_STATE, ALPHA_FX3_START_SCK0, FX3_GPIF_SWITCH_TIMEOUT);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(6, "\n\rCyFx3UvcAppStart:CyU3PGpifSMSwitch Err = 0x%x", status);
        return status;
    }

    CyU3PGpioSetValue(SLRD_GPIO, CyTrue);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStart: App started\n\r");
    return CY_U3P_SUCCESS;
}

/* This function stops the video streaming. It is called from the USB event
 * handler, when there is a reset / disconnect or SET_INTERFACE for alternate
 * interface 0 in case of ischronous implementation or when a Clear Feature (Halt)
 * request is recieved (in case of bulk only implementation).
 */
void
CyFx3UvcAppStop(
        CyBool_t isStillCapture)
{
    glIsStreamingStarted = CyFalse;
    CyU3PGpioSetValue(SLRD_GPIO, CyFalse);

#if FX3_DEBUG_ENABLE_LEVEL2
    uint8_t SMState = 0;

    CyU3PGpifGetSMState(&SMState);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcAppStop:SMState = 0x%x", SMState);
#endif

    /* Pause the GPIF interface*/
    CyU3PGpifDisable(CyFalse);

#ifdef FRAME_TIMER_ENABLE
    CyU3PTimerStop(&FrameTimer);
#endif

    /* Update the flag so that the application thread is notified of this. */
    glIsApplnActive = CyFalse;

    CyU3PUsbSetEpNak(FX3_EP_BULK_VIDEO_1, CyTrue);
    CyU3PBusyWait(FX3_EP_BUSY_WAIT);

    /* Abort and destroy the video streaming channel */
    /* Reset the channel: Set to DSCR chain starting point in PORD/CONS SCKT; set DSCR_SIZE field in DSCR memory*/
    CyU3PDmaMultiChannelReset(&glChHandleUVCStream);

    /* Flush the endpoint memory */
    CyU3PUsbFlushEp(FX3_EP_BULK_VIDEO_1);
    CyU3PUsbSetEpNak(FX3_EP_BULK_VIDEO_1, CyFalse);
    CyU3PBusyWait(FX3_EP_BUSY_WAIT);

    /* Clear the stall condition and sequence numbers if ClearFeature. */
    if (glIsClearFeature)
    {
        CyU3PUsbStall(FX3_EP_BULK_VIDEO_1, CyFalse, CyTrue);
        glIsClearFeature = CyFalse;
    }
    glDMATxCount = 0;
    glDmaDone = 0;

    glFrameEnd = CyFalse;
    /* Enable USB 3.0 LPM */
    CyU3PUsbLPMEnable();
    glLpmDisable = CyFalse;
    CyU3PDebugPrint(4,"\r\n%s",__FUNCTION__);

}

/* GpifCB callback function is invoked when FV triggers GPIF interrupt */
void CyFx3UvcAppGpifCB(CyU3PGpifEventType event, uint8_t currentState)
{
    /* Handle interrupt from the State Machine */
    if (event == CYU3P_GPIF_EVT_SM_INTERRUPT)
    {
        switch (currentState)
        {
            default:
                break;
        }
    }
}

/* This is the Callback function to handle the USB Events */
static void
CyFx3UvcAppUSBEventCB(
        CyU3PUsbEventType_t evtype, /* Event type */
        uint16_t evdata /* Event data */)
{
    uint8_t interface = 0, altSetting = 0;

    switch (evtype)
    {
    case  CY_U3P_USB_EVENT_SS_COMP_ENTRY:
#ifdef TYPE_C_ENABLED
        U3_EVENT = CyTrue;
        U3_TRUE = CyFalse;
        CyU3PDebugPrint(5, "\n\rCyFx3UvcAppUSBEventCB: SS_COMP_ENTRY Event");
#endif
        break;

    case  CY_U3P_USB_EVENT_USB3_LNKFAIL:
#ifdef TYPE_C_ENABLED
        U3_EVENT = CyTrue;
        U3_TRUE = CyFalse;
        CyU3PDebugPrint(5, "\n\rCyFx3UvcAppUSBEventCB: USB3_LNKFAIL Event");
#endif
        break;

    case CY_U3P_USB_EVENT_SUSPEND:
        /* Suspend the device with Wake On Bus Activity set */
        /* Set timer expire period to ISP_LOW_POWER_ENTRY_TIMEOUT so that
           FX3 does not de-initialize ISP on every Clear feature request */
//        CyU3PTimerModify(&UvcTimer, ISP_LOW_POWER_ENTRY_TIMEOUT, 0);
        glReceivedSuspEvent = CyTrue;
        break;

    case CY_U3P_USB_EVENT_SETINTF:
        /* Start the video streamer application if the
         * interface requested was 1. If not, stop the
         * streamer. */
        interface = CY_U3P_GET_MSB(evdata);
        altSetting = CY_U3P_GET_LSB(evdata);

#ifdef FX3_DEBUG_ENABLE_LEVEL1
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBEventCB: IF = %d, ALT = %d", interface, altSetting);
#endif

        if(altSetting == 0 && (glIsApplnActive) )
        {
            glIsStreamingStarted = CyFalse;

            /* Make sure that the endpoint toggles/sequence numbers are cleared. */
            glIsClearFeature = CyTrue;

            /* For ISOC we are receiving SET interface event to stop streaming video data.
             * Whereas for BULK we get Clear feature request. */
            glRstDmaOnEPClearFeature = CyTrue;

#ifdef FX3_DEBUG_ENABLE_LEVEL2
            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBEventCB:Calling AppStop");
#endif
            CyFx3UvcAppStop(CyFalse);
        }

        if ((altSetting == FX3_UVC_STREAM_INTERFACE) && (interface == FX3_UVC_STREAM_INTERFACE_1))
        {
            CyFx3UvcAppStart(FX3_EP_BULK_VIDEO_1);
        }
        break;

    case CY_U3P_USB_EVENT_SETCONF:
        if (CyU3PUsbGetSpeed() == CY_U3P_SUPER_SPEED)
        {
            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBEventCB: Detected SS USB Connection");
        }
        else if (CyU3PUsbGetSpeed() == CY_U3P_HIGH_SPEED)
        {
            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBEventCB: Detected HS USB Connection");
        }
        break;

        /* Intentional Fall-through all cases */
    case CY_U3P_USB_EVENT_RESET:
    case CY_U3P_USB_EVENT_DISCONNECT:
    case CY_U3P_USB_EVENT_CONNECT:
        glIsStreamingStarted = CyFalse;
        if (evtype == CY_U3P_USB_EVENT_SETCONF)
            glIsConfigured = CyTrue;
        else
            glIsConfigured = CyFalse;

        /* Stop the video streamer application and enable LPM. */
        CyU3PUsbLPMEnable();
        if (glIsApplnActive)
        {
#ifdef FX3_DEBUG_ENABLE_LEVEL2
            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBEventCB:Calling AppStop");
#endif
            CyFx3UvcAppStop(CyFalse);
        }
        break;

    case CY_U3P_USB_EVENT_EP_UNDERRUN:
        CyU3PDebugPrint(4, "\n\rrCyFx3UvcAppUSBEventCB: USB Buffer Under-run Event ");
        break;
    default:
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBEventCB: Default USB evtype = %d", evtype);
        break;
    }
}

/* Callback for LPM requests. Always return true to allow host to transition device
 * into required LPM state U1/U2/U3. When data trasmission is active LPM management
 * is explicitly disabled to prevent data transmission errors.
 */
static CyBool_t CyFx3UvcAppLPMRqtCB(
        CyU3PUsbLinkPowerMode link_mode /*USB 3.0 linkmode requested by Host */)
{
    return CyTrue;
}

static void
CyFx3UvcAppHandleSetCurReq(
        uint16_t wValue,
        uint16_t wIndex)
{
    CyU3PReturnStatus_t status;
    uint16_t readCount = 0;
    uint32_t frameInterval = 0;

    if (glIsApplnActive)
    {
#ifdef FX3_DEBUG_ENABLE_LEVEL2
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppHandleSetCurReq: Calling CyFx3UvcAppStop");
#endif
        CyFx3UvcAppStop(CyFalse);
    }

    /* Get the UVC probe/commit control data from EP0 */
    status = CyU3PUsbGetEP0Data(FX3_UVC_MAX_PROBE_SETTING_ALIGNED, glCommitCtrl, &readCount);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppHandleSetCurReq:GetEP0Data Err = 0x%x", status);
        return;
    }

    /* Check the read count. Expecting a count of FX3_UVC_MAX_PROBE_SETTING bytes. */
    if (readCount > (uint16_t)FX3_UVC_MAX_PROBE_SETTING)
    {
        CyU3PDebugPrint(4, "\n\rrCyFx3UvcAppHandleSetCurReq:Invalid Rqt Len");
        return;
    }

    /* Set Probe Control */
    if (wValue == FX3_UVC_VS_PROBE_CONTROL)
    {
        glCurrentVideoFormatIndex = glCommitCtrl[2];
        glCurrentVideoFrameIndex = glCommitCtrl[3];
        glCurrentFrameInterval = ((uint32_t)glCommitCtrl[7] << 24) | ((uint32_t)glCommitCtrl[6] << 16)
                                                                               | ((uint32_t)glCommitCtrl[5] << 8) | ((uint32_t)glCommitCtrl[4] << 0);
#ifdef FX3_DEBUG_ENABLE_LEVEL1
        CyU3PDebugPrint(4, "\n\rrCyFx3UvcAppHandleSetCurReq: Set Probe Ctrl: Video Format Index= %d Frame Index= %d Frame Interval = %d\n\r",
                glCommitCtrl[2], glCommitCtrl[3], glCurrentFrameInterval);
#endif
    }
    else
    {
        if (wValue == FX3_UVC_VS_COMMIT_CONTROL)
        {
            frameInterval = ((uint32_t)glCommitCtrl[7] << 24) | ((uint32_t)glCommitCtrl[6] << 16)
                                                                         | ((uint32_t)glCommitCtrl[5] << 8) | ((uint32_t)glCommitCtrl[4] << 0);
#ifdef FX3_DEBUG_ENABLE_LEVEL1
            CyU3PDebugPrint(4, "\n\rrCyFx3UvcAppHandleSetCurReq: Set Commit Ctrl: Video Format Index= %d Frame Index= %d Frame Interval = %d\n\r",
                    glCommitCtrl[2], glCommitCtrl[3], frameInterval);
#endif


            switch (CY_U3P_GET_LSB(wIndex))
            {
            case FX3_UVC_STREAM_INTERFACE_1:
            {
                CyU3PThreadSleep(900);

                CyFx3UvcAppSetVideoResolution(glCommitCtrl[2], glCommitCtrl[3], frameInterval, 0);
#ifndef USE_ISOCHRONOUS_ENDPOINT
                CyFx3UvcAppStart(FX3_EP_BULK_VIDEO_1);
#endif
            }
            break;
            }
        }
    }

}

/* Callback to handle the USB Setup Requests and UVC Class events */
static CyBool_t CyFx3UvcAppUSBSetupCB(
        uint32_t setupdat0, /* SETUP Data 0 */
        uint32_t setupdat1 /* SETUP Data 1 */)
{
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    uint8_t bRequest, bType, bRType, bTarget;
    uint16_t wValue, wIndex, wLength;;
    uint8_t ep0Buf[512];
    CyBool_t isHandled = CyFalse;
    uint8_t *probeCtrlStruct = 0;
    uint16_t readCount = 0;

#ifdef STILL_CAPTURE_ENABLE
    uint32_t eventFlagSC;
#endif

    /* Decode the fields from the setup request. */
    bRType   = (setupdat0 & CY_U3P_USB_REQUEST_TYPE_MASK);
    bType    = (bRType & CY_U3P_USB_TYPE_MASK);
    bTarget  = (bRType & CY_U3P_USB_TARGET_MASK);
    bRequest = ((setupdat0 & CY_U3P_USB_REQUEST_MASK) >> CY_U3P_USB_REQUEST_POS);
    wValue   = ((setupdat0 & CY_U3P_USB_VALUE_MASK) >> CY_U3P_USB_VALUE_POS);
    wIndex   = ((setupdat1 & CY_U3P_USB_INDEX_MASK) >> CY_U3P_USB_INDEX_POS);
    wLength  = ((setupdat1 & CY_U3P_USB_LENGTH_MASK)  >> CY_U3P_USB_LENGTH_POS);

    /* Make sure you clear the timer once you receive a USB packet from Host, i.e.
     * Timeout between receiving suspend event and acknowledging it through Timeout */
    if(glReceivedSuspEvent)
    {
        CyU3PDebugPrint(0,"\r\nRcvd suspend evt");

        CyU3PTimerStop(&FrameTimer);
        glReceivedSuspEvent = CyFalse;
    }

    CyU3PDebugPrint(10, "\n\rCyFx3UvcAppUSBSetupCB: bRType = 0x%x, bRequest = 0x%x, wValue = 0x%x, wIndex = 0x%x, wLength= 0x%x",
            bRType, bRequest, wValue, wIndex, wLength);

    /* Handle Microsoft OS String Descriptor request. */
    if ((bTarget == CY_U3P_USB_TARGET_DEVICE) && (bRequest == CY_U3P_USB_SC_GET_DESCRIPTOR) &&
            (wValue == ((CY_U3P_USB_STRING_DESCR << 8) | CY_U3P_USB_MS_STRING_DSCR)))
    {
        if(wLength > CyFx3UsbMsOSStringDscr[0])
            wLength = CyFx3UsbMsOSStringDscr[0];

        CyU3PDebugPrint(6, "\n\r\n\rCyFx3UvcAppUSBSetupCB: Received OS String Descriptor request, wLength = 0x%x\n\r", wLength);
        CyU3PUsbSendEP0Data (wLength, (uint8_t *)CyFx3UsbMsOSStringDscr);
        isHandled = CyTrue;
    }

    /* Handle Microsoft OS Extended Descriptor request. */
    if ((bRType == CY_U3P_USB_GET_MS_EXT_PROP_DSCR) && (wIndex == CY_U3P_USB_MS_EXT_PROP_TYPE))
    {
        if(wLength == CY_U3P_USB_MS_EXT_PROP_HDR_SIZE || wLength == CY_U3P_USB_MS_EXT_PROP_DSCR_SIZE)
        {
            CyU3PUsbSendEP0Data (wLength, (uint8_t *)CyFx3UsbMsOSExtPropDscr);
#ifdef FX3_DEBUG_ENABLE_LEVEL1
            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: OS Descriptor request, wValue = 0x%x, wIndex = 0x%x, wLength= 0x%x", wValue, wIndex, wLength);
#endif
        }
    }

    /* ClearFeature(Endpoint_Halt) received on the Streaming Endpoint. Stop Streaming */
    if ((bTarget == CY_U3P_USB_TARGET_ENDPT) && (bRequest == CY_U3P_USB_SC_CLEAR_FEATURE)
            && (wIndex == FX3_EP_BULK_VIDEO_1) && (wValue == CY_U3P_USBX_FS_EP_HALT))
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: Received EP Clear Feature");
        if (glIsApplnActive)
        {
            glRstDmaOnEPClearFeature = CyTrue;
            glIsClearFeature = CyTrue;

            CyFx3UvcAppStop(CyFalse);
            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: rStopping Video Streaming\n\r");
        }

        return CyFalse;
    }

    if (bRType == CY_U3P_USB_GS_DEVICE)
    {
        /* Make sure that we bring the link back to U0, so that the ERDY can be sent. */
        if (CyU3PUsbGetSpeed() == CY_U3P_SUPER_SPEED)
            CyU3PUsbSetLinkPowerState(CyU3PUsbLPM_U0);
    }

    if ((bTarget == CY_U3P_USB_TARGET_INTF) &&
       ((bRequest == CY_U3P_USB_SC_SET_FEATURE) ||
        (bRequest == CY_U3P_USB_SC_CLEAR_FEATURE)) && (wValue == 0))
    {
#ifdef FX3_DEBUG_ENABLE_LEVEL2
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: SET_FTR %d::%d", glIsApplnActive_1, glIsConfigured);
#endif
        if (glIsConfigured)
        {
            CyU3PUsbAckSetup();
        }
        else
        {
            CyU3PUsbStall(0, CyTrue, CyFalse);
        }
        return CyTrue;
    }

    if ((bRequest == CY_U3P_USB_SC_GET_STATUS) && (bTarget == CY_U3P_USB_TARGET_INTF))
    {
        /* We support only interface 0. */
        if (wIndex == 0)
        {
            ep0Buf[0] = 0;
            ep0Buf[1] = 0;
            CyU3PUsbSendEP0Data(0x02, ep0Buf);
        }
        else
            CyU3PUsbStall(0, CyTrue, CyFalse);
        return CyTrue;
    }

    /* Check for UVC Class Requests */
    if (bType == CY_U3P_USB_CLASS_RQT)
    {
        /* Requests to the Video Streaming Interface */
        if (CY_U3P_GET_LSB(wIndex) == FX3_UVC_STREAM_INTERFACE_1)
        {
            if ((wValue == FX3_UVC_VS_PROBE_CONTROL) || (wValue == FX3_UVC_VS_COMMIT_CONTROL))
            {
                switch (bRequest)
                {
                case FX3_USB_UVC_GET_INFO_REQ:
                {
                    ep0Buf[0] = 3;
                    CyU3PUsbSendEP0Data(1, (uint8_t *)ep0Buf);
                    isHandled = CyTrue;
                }
                break;

                case FX3_USB_UVC_GET_LEN_REQ:
                {
                    ep0Buf[0] = FX3_UVC_MAX_PROBE_SETTING;
                    CyU3PUsbSendEP0Data(1, (uint8_t *)ep0Buf);
                    isHandled = CyTrue;
                }
                break;

                case FX3_USB_UVC_GET_CUR_REQ:
                case FX3_USB_UVC_GET_MIN_REQ:
                case FX3_USB_UVC_GET_MAX_REQ:
                case FX3_USB_UVC_GET_DEF_REQ:
                {
                    /* Host requests for probe data of 34 bytes (UVC 1.1) or 26 Bytes (UVC1.0). Send it over EP0. */
#ifdef FX3_DEBUG_ENABLE_LEVEL1
                    CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: Get Probe Cntrl: Video Format Index= %d Frame Index= %d Frame Interval = %d",
                            glCurrentVideoFormatIndex, glCurrentVideoFrameIndex, glCurrentFrameInterval);
#endif
                    probeCtrlStruct = CyFx3UvcAppGetVideoProbeCtrlStruct(CyU3PUsbGetSpeed(), glCurrentVideoFormatIndex,
                            glCurrentVideoFrameIndex, glCurrentFrameInterval);
                    if (probeCtrlStruct != 0)
                    {
                        CyU3PMemCopy(glProbeCtrl, (uint8_t *)probeCtrlStruct, FX3_UVC_MAX_PROBE_SETTING);

                        status = CyU3PUsbSendEP0Data(FX3_UVC_MAX_PROBE_SETTING, glProbeCtrl);
                        if (status != CY_U3P_SUCCESS)
                        {
                            CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: GET_CUR: CyU3PUsbSendEP0Data Err = 0x%x", status);
                        }
                    }
                    else
                    {
                        CyU3PUsbStall(0, CyTrue, CyFalse);
                    }

                    isHandled = CyTrue;
                }
                break;

                case FX3_USB_UVC_SET_CUR_REQ:
                {
                    /* Make sure that the endpoint is ready to receive clear feature command and reset DMA accordingly */
                    glRstDmaOnEPClearFeature = CyFalse;


                    CyFx3UvcAppHandleSetCurReq(wValue, wIndex);

                    isHandled = CyTrue;
                }
                break;

                default:
                    isHandled = CyFalse;
                    break;
                }
            }
#ifdef STILL_CAPTURE_ENABLE
            else if ((wValue == FX3_UVC_VS_STILL_PROBE_CONTROL) || (wValue == FX3_UVC_VS_STILL_COMMIT_CONTROL))
            {
                switch (bRequest)
                {
                case FX3_USB_UVC_GET_CUR_REQ:
                case FX3_USB_UVC_GET_MIN_REQ:
                case FX3_USB_UVC_GET_MAX_REQ:
                {
                    CyU3PDebugPrint(6,"\n\rCyFx3UvcAppUSBSetupCB: Get Probe Ctrl: Still Format Index = %d Frame Index = %d", glCurrentStillFormatIndex, glCurrentStillFrameIndex);
                    glStillProbeCtrl[0] = glCurrentStillFormatIndex;
                    glStillProbeCtrl[1] = glCurrentStillFrameIndex;

                    status = CyU3PUsbSendEP0Data(FX3_UVC_MAX_STILL_PROBE_SETTING, (uint8_t*)glStillProbeCtrl);
                    if (status != CY_U3P_SUCCESS)
                        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: Still Capture CyU3PUsbSendEP0Data Failed 0x%x", status);
                }
                break;

                case FX3_USB_UVC_SET_CUR_REQ:
                {
                    /* Get the UVC probe/commit control data from EP0 */
                    status = CyU3PUsbGetEP0Data(16, glStillCommitCtrl, &readCount);
                    if (status != CY_U3P_SUCCESS)
                        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppUSBSetupCB: Still CyU3PUsbGetEP0Data Err = 0x%x", status);

                    glCurrentStillFormatIndex = glStillCommitCtrl[0];
                    glCurrentStillFrameIndex = glStillCommitCtrl[1];
                    CyU3PDebugPrint(6,"\n\rCyFx3UvcAppUSBSetupCB: Set Probe Ctrl: Still Format Index = %d Frame Index = %d", glCurrentStillFormatIndex, glCurrentStillFrameIndex);
                }
                break;
                }
                return CyTrue;
            }
            else if (wValue == FX3_UVC_VS_STILL_IMAGE_TRIGGER_CONTROL)
            {

                glStillFlag = CyTrue;
                status = CyU3PUsbGetEP0Data(16, glStillReqBuff, &readCount);
                if (status != CY_U3P_SUCCESS)
                {
                    CyU3PDebugPrint(6, "\n\rCyFx3UvcAppUSBSetupCB: Still CyU3PUsbGetEP0Data Err = 0x%x", status);
                }
                CyU3PDebugPrint(6, "\n\r\r\nCyFx3UvcAppUSBSetupCB: Received Still Capture Event");
                CyU3PDebugPrint(6, "\n\rWait for current frame complete");
                /*Waiting for the current frame to end*/
                CyU3PEventGet(&glStillImageEvent, FX3_STILL_IMAGE_EVENT, CYU3P_EVENT_OR_CLEAR, &eventFlagSC, CYU3P_WAIT_FOREVER);
                CyFx3UvcAppStop(CyFalse);
                glStillHeader = CyTrue;
                CyFx3UvcAppSetStillResolution (glCurrentStillFormatIndex, glCurrentStillFrameIndex);
                CyFx3UvcAppStart(FX3_EP_BULK_VIDEO_1);
                /*Waiting for the current frame to end*/

                CyU3PEventGet(&glStillImageEvent, FX3_STILL_IMAGE_EVENT, CYU3P_EVENT_OR_CLEAR, &eventFlagSC, CYU3P_WAIT_FOREVER);
                CyU3PThreadSleep(30);
                CyFx3UvcAppStop(CyTrue);
                glStillHeader =CyFalse;
                glStillFlag = CyFalse;
                CyFx3UvcAppSetVideoResolution(glCurrentVideoFormatIndex, glCurrentVideoFrameIndex, glCurrentFrameInterval, 0);
                CyFx3UvcAppStart(FX3_EP_BULK_VIDEO_1);
                return CyTrue;
            }
#endif
        }

        /* Request addressed to the Video Control Interface */
        else if (CY_U3P_GET_LSB(wIndex) == FX3_UVC_CONTROL_INTERFACE_1)
        {

            /* Respond to VC_REQUEST_ERROR_CODE_CONTROL and stall every other request as this example does
       not support any of the Video Control features */
#ifdef VIDEO_CONTROL_INTERFACE
            status = CyFx3UvcHandleVCInterfaceReq(wIndex, wValue, bRequest);
                if (status != CY_U3P_SUCCESS)
                {
                    isHandled = CyFalse;
                }
                else
                    isHandled = CyTrue;
#endif
        }
    }

    return isHandled;
}


void
CyFx3SPIDmaCallback(
        CyU3PDmaChannel     *ChHandle,
        CyU3PDmaCbType_t    type,
        CyU3PDmaCBInput_t   *input
)
{

}

void CyFxGpioInit(void)
{
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    CyU3PGpioClock_t gpioClock;
    CyU3PI2cConfig_t i2cConfig;

    status = CyU3PI2cInit ();
    if (status != CY_U3P_SUCCESS)
    {
        return;
    }

    /* Start the I2C master block. The bit rate is set at 100KHz.
     * The data transfer is done via DMA. */
    CyU3PMemSet ((uint8_t *)&i2cConfig, 0, sizeof(i2cConfig));
    i2cConfig.bitRate    = 100000;
    i2cConfig.busTimeout = 0xFFFFFFFF;
    i2cConfig.dmaTimeout = 0xFFFF;
    i2cConfig.isDma      = CyFalse;

    status = CyU3PI2cSetConfig (&i2cConfig, NULL);
    if (status != CY_U3P_SUCCESS)
    {
        return;
    }

    /* Steps to configure GPIO(s)
        1.Configure and initialize clock for GPIO Module (CyU3PGpioClock_t)
        2.Override IOs to GPIOs if needed
        3.Configure each GPIO using CyU3PGpioSetSimpleConfig()
    */


    /*Initializing GPIO Block*/
    gpioClock.fastClkDiv = 16;
    gpioClock.slowClkDiv = 64;
    gpioClock.simpleDiv = CY_U3P_GPIO_SIMPLE_DIV_BY_2;
    gpioClock.clkSrc = CY_U3P_SYS_CLK;
    gpioClock.halfDiv = 0;

    status = CyU3PGpioInit(&gpioClock, NULL);
    if (status != CY_U3P_SUCCESS)
    {
        /* Error Handling */
        CyU3PDebugPrint(4, "CyU3PGpioInit failed, error code = %d\n", status);
        CyFx3UvcAppErrorHandler(status);
    }

    CyU3PDeviceGpioOverride(FX3_LED, CyTrue);
    CyU3PDeviceGpioOverride(PROG_N_GPIO, CyTrue);
    CyU3PDeviceGpioOverride(INIT_N_GPIO, CyTrue);
    CyU3PDeviceGpioOverride(DONE_GPIO, CyTrue);

    CyU3PDeviceGpioOverride(SLOE_GPIO, CyTrue);
    CyU3PDeviceGpioOverride(SLRD_GPIO, CyTrue);
    CyU3PDeviceGpioOverride(MUX_SEL_GPIO, CyTrue);

                                     /* outValue, driveLowEn, driveHighEn, inputEn, interruptMode */
    CyU3PGpioSimpleConfig_t testConf = { CyTrue, CyTrue, CyTrue, CyFalse, CY_U3P_GPIO_NO_INTR };

    status = CyU3PGpioSetSimpleConfig(FPGA_SSN, &testConf);
    status = CyU3PGpioSetSimpleConfig(FX3_LED, &testConf);
    status = CyU3PGpioSetSimpleConfig(PROG_N_GPIO, &testConf);

    testConf.outValue = CyFalse;
    status = CyU3PGpioSetSimpleConfig(SLOE_GPIO, &testConf);
    status = CyU3PGpioSetSimpleConfig(SLRD_GPIO, &testConf);
    status = CyU3PGpioSetSimpleConfig(MUX_SEL_GPIO, &testConf);

    testConf.driveHighEn = CyFalse;
    testConf.driveLowEn = CyFalse;
    testConf.inputEn = CyTrue;
    status = CyU3PGpioSetSimpleConfig(DONE_GPIO, &testConf);
    status = CyU3PGpioSetSimpleConfig(INIT_N_GPIO, &testConf);
    status = CyU3PGpioSetSimpleConfig(SW2_GPIO, &testConf);
    status = CyU3PGpioSetSimpleConfig(SW3_GPIO, &testConf);

}

uint32_t CyReadRegister(uint8_t cmd)
{
    uint8_t location[4] = { 0 };
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    uint8_t read_buffer[4];

    location[0] = cmd;

    CyU3PGpioSetValue(FPGA_SSN, CyFalse);
    status = CyU3PSpiTransmitWords(location, 4);
    if (status != 0)
    {
        CyU3PDebugPrint(4, "\n\rCyReadRegister:Sending %x cmd failed", cmd);
        CyU3PGpioSetValue(FPGA_SSN, CyTrue);
        return 0;
    }

    status = CyU3PSpiReceiveWords(read_buffer, 4);
    if (status != 0)
    {
        CyU3PDebugPrint(4, "\n\rCyReadRegister:Reading %x cmd value failed", cmd);
        CyU3PGpioSetValue(FPGA_SSN, CyTrue);
        return 0;
    }
    CyU3PGpioSetValue(FPGA_SSN, CyTrue);

    return ((read_buffer[0] << 24) + (read_buffer[1] << 16)
            + (read_buffer[2] << 8) + read_buffer[3]);

}

static CyU3PReturnStatus_t CyFx3UvcConfigureFpga(void)
{
    uint32_t status = 0;
    CyBool_t gpio_value = CyFalse;
    uint8_t location[4] = { 0 };
    uint32_t fpga_id;

    CyU3PGpioSimpleSetValue(PROG_N_GPIO, CyFalse);
    CyU3PBusyWait(1);
    CyU3PGpioSimpleSetValue(PROG_N_GPIO, CyTrue);

    while(gpio_value == CyFalse)
    {
        CyU3PGpioSimpleGetValue(INIT_N_GPIO, &gpio_value);
        CyU3PBusyWait(1000);
    }

    fpga_id = CyReadRegister(READ_ID_CMD);
    if (fpga_id != ECP5_5G_ID)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:Device ID mismatch 0x%x", fpga_id);
        return 1;
    }
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:Device ID 0x%x", fpga_id);
    CyU3PBusyWait(50);

    CySendSpiCmd(WRITE_EN_CMD);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:ISC ENABLE Command Sent");
    CyU3PThreadSleep(5);

    status = CyReadRegister(READ_STATUS_CMD);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:CyReadRegister Status =  0x%x", status);
    CyU3PThreadSleep(4);

    CySendSpiCmd(SRAM_ERASE);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:SRAM ERASE Command Sent");
    CyU3PThreadSleep(200);

    CySendSpiCmd(INIT_ADDR);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:INIT ADDR Command Sent");
    CyU3PBusyWait(50);

    location[0] = WRITE_BITS_CMD;
    CyU3PGpioSetValue(FPGA_SSN, CyFalse);
    status = CyU3PSpiTransmitWords(location, 4);
    if (status != 0)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:Sending 0x82 cmd failed");
    }
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:ISC BITSTREAM BURST Command Sent");

    CyWriteStreamCmd();
    CyU3PThreadSleep(10);

    status = CyReadRegister(READ_STATUS_CMD);
    CyU3PDebugPrint(4, "\n\rCyFx3UvcConfigureFpga:CyReadRegister Status = 0x%x", status);
    CyU3PBusyWait(50);

    CySendSpiCmd(WRITE_DIS_CMD);
    CyU3PThreadSleep(200);

    CySendSpiCmd(0xFF);

    gpio_value = CyFalse;
    while(gpio_value == CyFalse)
        CyU3PGpioSimpleGetValue(DONE_GPIO, &gpio_value);

    if (status == 0x200F00)
    {
        CyU3PGpioSetValue(FX3_LED, CyFalse);
        CyU3PThreadSleep(50);
        CyU3PGpioSetValue(SLOE_GPIO, CyTrue);
        CyU3PGpioSetValue(SLRD_GPIO, CyTrue);

    }

    return CY_U3P_SUCCESS;
}


/* This function initializes the USB Module, creates event group,
 sets the enumeration descriptors, configures the Endpoints and
 configures the DMA module for the UVC Application */
void CyFx3UvcAppInit(void)
{
    CyU3PEpConfig_t endPointConfig;
    CyU3PDmaMultiChannelConfig_t dmaCfg;
    CyU3PDmaChannelConfig_t dmaConfig;
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    CyFxGpioInit();

    /*Initializing SPI Block for FPGA Interface*/

    status = CyFx3SpiInit(CyFalse, CyFalse);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:SPIInit Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* Start the USB functionality */
    status = CyU3PUsbStart();
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:UsbStart Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* The fast enumeration is the easiest way to setup a USB connection,
     * where all enumeration phase is handled by the library. Only the
     * class / vendor requests need to be handled by the application. */
    CyU3PUsbRegisterSetupCallback(CyFx3UvcAppUSBSetupCB, CyTrue);

    /* Setup the callback to handle the USB events */
    CyU3PUsbRegisterEventCallback(CyFx3UvcAppUSBEventCB);

    /* Register a callback to handle LPM requests from the USB 3.0 host. */
    CyU3PUsbRegisterLPMRequestCallback(CyFx3UvcAppLPMRqtCB);

    /* Set the USB Enumeration descriptors */

    /* Super speed device descriptor. */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_DEVICE_DESCR, 0,(uint8_t *)CyFx3USB30DeviceDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_SS_Device_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* High speed device descriptor. */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_DEVICE_DESCR, 0,(uint8_t *)CyFx3USB20DeviceDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_HS_Device_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* BOS descriptor */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_BOS_DESCR, 0,(uint8_t *)CyFx3USBBOSDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_BOS_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* Device qualifier descriptor */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_DEVQUAL_DESCR, 0,(uint8_t *)CyFx3USBDeviceQualDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_DEVQUAL_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* Super speed configuration descriptor */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_CONFIG_DESCR, 0, (uint8_t *)CyFx3USBSSConfigDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_SS_CFG_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* High speed configuration descriptor */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_CONFIG_DESCR, 0,(uint8_t *)CyFx3USBHSConfigDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_HS_CFG_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* Full speed configuration descriptor */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_FS_CONFIG_DESCR, 0,(uint8_t *)CyFx3USBFSConfigDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_FS_CFG_Dscr Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* String descriptor 0 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 0, (uint8_t *)CyFx3USBStringLangIDDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr0 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* String descriptor 1 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 1, (uint8_t *)CyFx3USBManufactureDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr1 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* String descriptor 2 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 2, (uint8_t *)CyFx3USBProductDescr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr2 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }
    /* String descriptor 3 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 3, (uint8_t *)CyFx3USBConfigSSDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr3 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* String descriptor 4 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 4, (uint8_t *)CyFx3USBConfigHSDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr4 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* String descriptor 5 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 5, (uint8_t *)CyFx3USBConfigFSDscr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr5 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* String descriptor 1 */
    status = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 6, (uint8_t *)CyFx3USBProductDescr);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:Set_STRNG_Dscr6 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /* Control status interrupt endpoint configuration */
    endPointConfig.enable = CyTrue;
    endPointConfig.epType = CY_U3P_USB_EP_INTR;
    endPointConfig.pcktSize = FX3_EP_INTR_PACKET_SIZE;
    endPointConfig.isoPkts = 1;
    endPointConfig.burstLen = FX3_EP_INTR_BURST_LEN;

    status = CyU3PSetEpConfig(FX3_EP_CONTROL_STATUS_1, &endPointConfig);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:CyU3PSetEpConfig CtrlEp Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    CyU3PUsbFlushEp(FX3_EP_CONTROL_STATUS_1);
    CyU3PThreadSleep(10);

    CyU3PUsbSetEpNak(FX3_EP_CONTROL_STATUS_1, CyFalse);
    CyU3PBusyWait(FX3_EP_BUSY_WAIT);

#ifdef USE_ISOCHRONOUS_ENDPOINT
    /* Setup the ISO endpoint used for Video Streaming */
    endPointConfig.enable = CyTrue;
    endPointConfig.epType = CY_U3P_USB_EP_ISO;
    endPointConfig.isoPkts = FX3_EP_ISO_VIDEO_SS_MULT + 1;
    endPointConfig.streams = 0;
    endPointConfig.pcktSize = FX3_EP_BULK_VIDEO_PKT_SIZE;
    endPointConfig.burstLen = FX3_EP_ISO_VIDEO_SS_BURST + 1;

#else
    /* Setup the Bulk endpoint used for Video Streaming */
    endPointConfig.enable = CyTrue;
    endPointConfig.epType = CY_U3P_USB_EP_BULK;
    endPointConfig.isoPkts = 0;
    endPointConfig.streams = 0;

    /* Have to wait for SetConfig event to use CyU3PUsbGetSpeed() API to configure the EP Size correctly.
     * Other option is to configure the EP for Super Speed connection. SDK will take care of correcting
     * the size as per USB connection */
    endPointConfig.pcktSize = FX3_EP_BULK_VIDEO_PKT_SIZE;
    endPointConfig.burstLen = FX3_EP_BULK_SUPER_SPEED_BURST_LEN;
#endif

    status = CyU3PSetEpConfig(FX3_EP_BULK_VIDEO_1, &endPointConfig);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:CyU3PSetEpConfig FX3_EP_BULK_VIDEO_1 Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    CyU3PUsbEPSetBurstMode(FX3_EP_BULK_VIDEO_1, CyTrue);

    /* Flush the endpoint memory */
    CyU3PUsbFlushEp(FX3_EP_BULK_VIDEO_1);


    /* Configure MANUAL MANY TO ONE DMA channel for GPIF to USB transfer */
    CyU3PMemSet ((uint8_t *)&dmaCfg, 0, sizeof (dmaCfg));
    dmaCfg.size  = 42*1024;
    dmaCfg.count = 2;
    dmaCfg.validSckCount = 2;
    dmaCfg.prodSckId[0] = CY_U3P_PIB_SOCKET_0;
    dmaCfg.prodSckId[1] = CY_U3P_PIB_SOCKET_1;
    dmaCfg.consSckId[0] = FX3_EP_VIDEO_CONS_SOCKET_1;
    dmaCfg.dmaMode = CY_U3P_DMA_MODE_BYTE;
    dmaCfg.notification = CY_U3P_DMA_CB_PROD_EVENT | CY_U3P_DMA_CB_CONS_EVENT;
    dmaCfg.cb = CyFx3ApplnDmaCallback;
    dmaCfg.prodHeader = 0;
    dmaCfg.prodFooter = 0;
    dmaCfg.consHeader = 0;
    dmaCfg.prodAvailCount = 0;

    status = CyU3PDmaMultiChannelCreate(&glChHandleUVCStream, CY_U3P_DMA_TYPE_MANUAL_MANY_TO_ONE, &dmaCfg);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:DmaMultiChannelCreate glChHandleUVCStream Err = 0x%x", status);
        CyU3PDeviceReset(CyFalse);
    }

    /* Create a DMA MANUAL_OUT channel for status. */
    dmaConfig.size = 32;
    dmaConfig.count = 1;
    dmaConfig.prodSckId = CY_U3P_CPU_SOCKET_PROD;
    dmaConfig.consSckId = FX3_EP_STATUS_CONS_SOCKET_1;
    dmaConfig.dmaMode = CY_U3P_DMA_MODE_BYTE;
    dmaConfig.notification = 0;
    dmaConfig.cb = NULL;
    dmaConfig.prodHeader = 0;
    dmaConfig.prodFooter = 0;
    dmaConfig.consHeader = 0;
    dmaConfig.prodAvailCount = 0;

    /* Create a DMA channel to send out the data from Status INT endpoint for Front Camera */
    status = CyU3PDmaChannelCreate(&glChHandleStatusIntrFC, CY_U3P_DMA_TYPE_MANUAL_OUT, &dmaConfig);
    if (status != CY_U3P_SUCCESS)
    {
        /* Error handling */
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:DmaChannelCreate glChHandleStatusIntrFC Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }
    CyU3PThreadSleep(10);

    status = CyU3PDmaChannelReset(&glChHandleStatusIntrFC);
    if (status != CY_U3P_SUCCESS)
    {
        /* Error handling */
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:CyU3PDmaChannelReset glChHandleStatusIntrFC Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    status = CyU3PDmaChannelSetXfer(&glChHandleStatusIntrFC, 0);
    if (status != CY_U3P_SUCCESS)
    {
        /* Error handling */
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:CyU3PDmaChannelSetXfer glChHandleStatusIntrFC Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }

    /*Update the UVC Extension Structure fields with the Firmware Version and Build Date of FX3 Firmware*/
    CyFx3GetFirmwareVersion();
    CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit: Cypress FX3 Firmware Version = %d.%d, Build Date = %d/%d/%d", glFx3FwVer[0], glFx3FwVer[1], glFx3FwVer[3], glFx3FwVer[2], glFx3FwVer[4]);
    CyU3PThreadSleep(10);

    /* Configure the FPGA */
    CyFx3UvcConfigureFpga();

    /* Re-configure the IO matrix to de-init SPI and init UART */
    CyFx3UvcIoMatrixReInit();

    /* Initialize and load the GPIF state machine */
    CyFx3UvcInitGpifSm();

    /* Wait for ISP to initialize and then write sensor configuration file */
    CyFx3_Sensor_Reg_I2c_Write_Bulk();

    /* Connect the USB pins and enable super speed operation */
#ifdef TYPE_C_ENABLED
    CyU3PThreadSleep(200);

    status = CyU3PUsbControlUsb2Support(CyFalse);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4,"\n\rCyFx3UvcAppInit:Disabling USB2 failed. Stopping Execution");
        while(1);
    }
    CyU3PThreadSleep(20);

    status = CyU3PConnectState(CyTrue, CyTrue);
    CyU3PThreadSleep(50);

    if(CyU3PUsbGetSpeed() != CY_U3P_SUPER_SPEED)
    {
        CyU3PDebugPrint(5,"\n\rCyFx3UvcAppInit:Not Super speed enumeration with MUX_SEL low. Fllipping Mux to try again");
        CyU3PConnectState(CyFalse, CyFalse);
        CyU3PGpioSetValue(MUX_SEL_GPIO, CyTrue);
        CyU3PUsbControlUsb2Support(CyTrue);
        CyU3PThreadSleep(20);
        status = CyU3PConnectState(CyTrue, CyTrue);
        CyU3PThreadSleep(50);
        if (status != CY_U3P_SUCCESS)
        {
            CyU3PDebugPrint (4, "\n\rCyFx3UvcAppInit:ConnectState Err = 0x%x", status);
            CyFx3UvcAppErrorHandler(status);
        }

    }
    else
    {
        CyU3PDebugPrint(5,"\n\rCyFx3UvcAppInit:Super speed enumeration with MUX_SEL low");
    }
#else
    status = CyU3PConnectState(CyTrue, CyTrue);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint(4, "\n\rCyFx3UvcAppInit:ConnectState Err = 0x%x", status);
        CyFx3UvcAppErrorHandler(status);
    }
#endif
    CyU3PThreadSleep(10);

#ifdef FRAME_TIMER_ENABLE
    CyU3PTimerCreate(&FrameTimer, CyFx3UvcAppProgressTimer, 0, glFrameTimerPeriod, 0, CYU3P_NO_ACTIVATE);
#endif

    CyU3PDebugPrint(4,"\n\rFX3 UVC App init complete");

}

/* This function initializes the debug module for the UVC application */
void CyFx3UvcAppDebugInit(void)
{
    CyU3PUartConfig_t uartConfig;
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    /* Errors in starting up the UART and enabling it for debug are not fatal errors.
     * Also, we cannot use DebugPrint until the debug module has been successfully initialized.
     */

    /* Initialize the UART for printing debug messages */
    status = CyU3PUartInit();
    if (status != CY_U3P_SUCCESS)
    {
        return;
    }

    /* Set UART Configuration */
    uartConfig.baudRate = CY_U3P_UART_BAUDRATE_115200;
    uartConfig.stopBit  = CY_U3P_UART_ONE_STOP_BIT;
    uartConfig.parity   = CY_U3P_UART_NO_PARITY;
    uartConfig.txEnable = CyTrue;
    uartConfig.rxEnable = CyFalse;
    uartConfig.flowCtrl = CyFalse;
    uartConfig.isDma    = CyTrue;

    /* Set the UART configuration */
    status = CyU3PUartSetConfig(&uartConfig, NULL);
    if (status != CY_U3P_SUCCESS)
    {
        return;
    }

    /* Set the UART transfer */
    status = CyU3PUartTxSetBlockXfer(0xFFFFFFFF);
    if (status != CY_U3P_SUCCESS)
    {
        return;
    }

    /* Initialize the debug application */
    status = CyU3PDebugInit(CY_U3P_LPP_SOCKET_UART_CONS, DEBUG_TRACE_LEVEL);
    if (status != CY_U3P_SUCCESS)
    {
        return;
    }

    CyU3PDebugPreamble(CyFalse);
}

CyU3PReturnStatus_t CyFx3UvcHandleDmaReset()
{
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;


    CyU3PThreadSleep(10);

#ifdef PRINT_FRAME_INFO
    /*Collect the frame information*/
    glFrameCount++;
    glPartialBufSize = 0;
    glConsCount = glCons;
    glProdCount = glProd;

    CyU3PDebugPrint(4,"\n\rCyFx3UvcHandleDmaReset:Prod = %d Cons = %d  Prtl_Sz = %d Frm_Sz = %d B", glProdCount, glConsCount, glPartialBufSize, ((glProdCount*FX3_UVC_DATA_BUF_SIZE)+glPartialBufSize));
    glProd = 0;
    glCons = 0;
    glProdCount = 0;
    glConsCount = 0;
    glPartialBufSize = 0;
#endif
    CyU3PDebugPrint(4, "\n\rCyFx3UvcHandleDmaReset:Processing DMA reset Event");

    /* We may get a commit buffer failure if host has issued stop streaming command
     * and FX3 already has data in its buffer. It will try to commit it and get failure. */
    if (glDmaResetFlag == 0)
    {
        CyU3PDebugPrint(4, "\n\r\n\rCyFx3UvcHandleDmaReset:Commit Buffer Failure");
        glCommitBufferFailureCount++;
    }
#ifdef FRAME_TIMER_ENABLE
    else if (glDmaResetFlag == 1)
    {
        CyU3PDebugPrint(4, "\n\r\n\rCyFx3UvcHandleDmaReset:Video timer, Timer Period = %d",
                glFrameTimerPeriod);
    }
#endif
    else if(glDmaResetFlag == 2)
    {
        CyU3PDebugPrint(4, "\n\r\n\rCyFx3UvcHandleDmaReset:Get Buffer Failure");
    }

    CyFx3UvcAppStop(CyFalse);
    if ((!glRstDmaOnEPClearFeature) && (glCommitBufferFailureCount < CY_FX_COMMIT_BUFFER_FAIL_THRESHOLD))
    {
        CyFx3UvcAppStart(FX3_EP_BULK_VIDEO_1);
    }
    glDmaResetFlag = 0;

    CyU3PDebugPrint(4, "\n\rCyFx3UvcHandleDmaReset:Exiting DMA reset Event");
    return status;
}

CyU3PReturnStatus_t CyFxSpiWaitForStatus(void)
{
    uint8_t buf[2], rd_buf[2];
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    /* Wait for status response from SPI flash device. */
    do {
        buf[0] = 0x06; /* Write enable command. */

        CyU3PSpiSetSsnLine(CyFalse);
        status = CyU3PSpiTransmitWords(buf, 1);
        CyU3PSpiSetSsnLine(CyTrue);
        if (status != CY_U3P_SUCCESS) {
            CyU3PDebugPrint(2, "\n\rCyFxSpiWaitForStatus:SPI WR_ENABLE command failed");
            return status;
        }

        buf[0] = 0x05; /* Read status command */

        CyU3PSpiSetSsnLine(CyFalse);
        status = CyU3PSpiTransmitWords(buf, 1);
        if (status != CY_U3P_SUCCESS) {
            CyU3PDebugPrint(2, "\n\rCyFxSpiWaitForStatus:SPI READ_STATUS command failed");
            CyU3PSpiSetSsnLine(CyTrue);
            return status;
        }

        status = CyU3PSpiReceiveWords(rd_buf, 2);
        CyU3PSpiSetSsnLine(CyTrue);
        if (status != CY_U3P_SUCCESS) {
            CyU3PDebugPrint(2, "\n\rCyFxSpiWaitForStatus:SPI status read failed");
            return status;
        }

    } while ((rd_buf[0] & 1) || (!(rd_buf[0] & 0x2)));

    return CY_U3P_SUCCESS;
}


CyU3PReturnStatus_t CyFxSpiTransfer(uint16_t pageAddress, uint16_t byteCount, uint8_t *buffer, CyBool_t isRead)
{
    uint8_t location[4];
    uint32_t byteAddress = 0;
    uint16_t pageCount = (byteCount / glSpiPageSize);
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    if (byteCount == 0) {
        return CY_U3P_SUCCESS;
    }
    if ((byteCount % glSpiPageSize) != 0) {
        pageCount++;
    }

    byteAddress = pageAddress * glSpiPageSize;

    pageCount = 1;
    while (pageCount != 0) {
        location[1] = (byteAddress >> 16) & 0xFF; /* MS byte */
        location[2] = (byteAddress >> 8) & 0xFF;
        location[3] = byteAddress & 0xFF; /* LS byte */

        if (isRead)
        {
            location[0] = 0x03; /* Read command. */

            status = CyFxSpiWaitForStatus();
            if (status != CY_U3P_SUCCESS)
                return status;

            CyU3PSpiSetSsnLine(CyFalse);

            status = CyU3PSpiTransmitWords(location, 4);
            if (status != CY_U3P_SUCCESS) {
                CyU3PDebugPrint(2, "\n\rCyFxSpiTransfer:SPI READ command failed");
                CyU3PSpiSetSsnLine(CyTrue);
                return status;
            }

            status = CyU3PSpiReceiveWords(buffer, byteCount);
            if (status != CY_U3P_SUCCESS) {
                CyU3PSpiSetSsnLine(CyTrue);
                return status;
            }

            CyU3PSpiSetSsnLine(CyTrue);

        }
        else /* Write */
        {
            location[0] = 0x02; /* Write command */

            status = CyFxSpiWaitForStatus();
            if (status != CY_U3P_SUCCESS)
                return status;

            CyU3PSpiSetSsnLine(CyFalse);
            status = CyU3PSpiTransmitWords(location, 4);
            if (status != CY_U3P_SUCCESS) {
                CyU3PDebugPrint(2, "\n\rCyFxSpiTransfer:SPI WRITE command failed");
                CyU3PSpiSetSsnLine(CyTrue);
                return status;
            }

            status = CyU3PSpiTransmitWords(buffer, glSpiPageSize);
            if (status != CY_U3P_SUCCESS) {
                CyU3PSpiSetSsnLine(CyTrue);
                return status;
            }

            CyU3PSpiSetSsnLine(CyTrue);
        }

        /* Update the parameters */
        byteAddress += glSpiPageSize;
        buffer += glSpiPageSize;
        pageCount--;
    }

    return CY_U3P_SUCCESS;
}



void CySendSpiCmd(uint8_t cmd)
{
    uint8_t location[4] = { 0 };
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    uint8_t length = 4;

    location[0] = cmd;

    if (cmd == WRITE_EN_CMD)
        length = 3;

    if (cmd == WRITE_DIS_CMD)
        length = 3;

    if (cmd == SRAM_ERASE)
        location[1] = 0x01;

    if (cmd == 0xFF) {
        location[1] = 0xFF;
        location[2] = 0xFF;
        location[3] = 0xFF;
    }

    CyU3PGpioSetValue(FPGA_SSN, CyFalse);
    status = CyU3PSpiTransmitWords(location, length);
    if (status != 0)
    {
        CyU3PDebugPrint(4, "\n\rCySendSpiCmd:Sending %x cmd failed", cmd);
    }
    CyU3PGpioSetValue(FPGA_SSN, CyTrue);

}


void CyWriteStreamCmd(void)
{
    uint32_t pageAdd = START_PAGE_ADDR;
    uint32_t byteCount = 0;
    uint8_t data_buff[1024] = { 0 };
    CyU3PReturnStatus_t status = 0;
    uint16_t length = 0;
    uint8_t location[4];
    uint32_t byteAdd;

    byteAdd = 0x4FFFC;
    location[0] = 0x03;
    location[1] = (byteAdd >> 16) & 0xFF; /* MS byte */
    location[2] = (byteAdd >> 8) & 0xFF;
    location[3] = byteAdd & 0xFF; /* LS byte */

    CyU3PSpiSetSsnLine(CyFalse);
    status = CyU3PSpiTransmitWords(location, 4);
    status = CyU3PSpiReceiveWords((uint8_t *)&byteCount, 4);
    CyU3PSpiSetSsnLine(CyTrue);


    CyU3PGpioSetValue(FPGA_SSN, CyFalse);

    while (byteCount > 0)
    {
        if (byteCount > 1024)
            length = 1024;
        else
            length = byteCount;
        status = CyFxSpiTransfer(pageAdd, length, data_buff, CyTrue);

        CyU3PGpioSetValue(FPGA_SSN, CyFalse);
        status = CyU3PSpiTransmitWords(data_buff, length);
        if (status != 0)
        {
            CyU3PDebugPrint(4, "\n\rCyWriteStreamCmd:Sending bit-streams failed");
        }
        byteCount -= length;
        pageAdd += 4;

        if (pageAdd % 100 == 0)
        {
            toggle = !toggle;
            CyU3PGpioSetValue(FX3_LED, toggle);
        }

    }
    CyU3PGpioSetValue(FPGA_SSN, CyTrue);
}


void CyFx3UvcIoMatrixReInit(void)
{
    uvint32_t *ioMatrixRegAddr = (uint32_t*) 0xE0051008; /* Address of GCTL_IOMATRIX register*/
    uint32_t cfg = 0, cfg1 = 0;



    CyU3PSpiDeInit();

    cfg = *(ioMatrixRegAddr);
    cfg1 = cfg;

    cfg &= (0 << CY_U3P_S1_CFG_POS);	//Disable SPI interface

    cfg |= CY_U3P_S0_CFG;				//Enable 32-bit GPIF interface
    cfg |= (5 << CY_U3P_S1_CFG_POS);	//Enable UART and I2S

    *(ioMatrixRegAddr) = cfg;
    CyU3PBusyWait(50);

    CyU3PDebugPrint(4, "\n\rCyFx3UvcIoMatrixReInit:Cfg1 = 0x%x, Cfg = 0x%x", cfg1, cfg);
}



void CyFx3UvcInitGpifSm(void)
{
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    CyU3PPibClock_t pibClk = {4, CyFalse, CyFalse, CY_U3P_SYS_CLK};

    CyU3PGpioSetValue(SLOE_GPIO, CyFalse);
    CyU3PGpioSetValue(SLRD_GPIO, CyFalse);

    /* Initialize the PIB block. */
    status = CyU3PPibInit (CyTrue, &pibClk);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "\n\rCyFx3UvcInitGpifSm: CyU3PPibInit failed, error code = 0x%x", status);
    }else{
        CyU3PDebugPrint (4, "\n\rCyFx3UvcInitGpifSm: CyU3PPibInit Success");

    }
    /* Load and start the GPIF state machine. */
    status = CyU3PGpifLoad (&CyFxGpifConfig);
    if (status != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "\n\rCyFx3UvcInitGpifSm: CyU3PGpifLoad failed, error code = 0x%x", status);
    }else{
        CyU3PDebugPrint (4, "\n\rCyFx3UvcInitGpifSm: CyU3PGpifLoad Success");

    }
    CyU3PGpioSetValue(SLOE_GPIO, CyTrue);

}

/* Entry function for the UVC application thread. */
void CyFx3UvcAppThread_Entry(uint32_t input)
{
    uint16_t wakeReason;
    uint32_t eventFlag;
    CyU3PReturnStatus_t status;
    CyU3PUsbLinkPowerMode usb3mode;

#ifdef PRINT_FRAME_INFO
    uint32_t fps = 0;
#endif

    /* Initialize the Debug Module */
    CyFx3UvcAppDebugInit();

    /* Initialize the UVC Application */
    CyFx3UvcAppInit();

    uint32_t flagsOfInterest = FX3_DMA_RESET_EVENT | FX3_USB_SUSP_EVENT_FLAG | FX3_USB_ASYNC_VC_EVENT |
                               FX3_USB_VC_AE_STATUS_CHK_EVENT | FX3_USB_VC_AF_STATUS_CHK_EVENT | FX3_USB_VC_AWB_STATUS_CHK_EVENT;

    for (;;)
    {
        eventFlag = 0;

#ifdef PRINT_FRAME_INFO
        if (glPrintFlag)
        {
            glPrintFlag = 0;
            /*For video streaming application of higher FPS refrain from using this debug print or try to reduce the print information*/
            CyU3PDebugPrint(4,"\n\rProd = %d Cons = %d Prtl_Sz = %d Frm_Sz = %d B", glProdCount, glConsCount, glPartialBufSize, glFrameSize);
#if FPS
            if (glFpsFlag)
            {
                glFpsFlag = 0;
                if(glPeriod)
                {
                    fps = 60000 / glPeriod; /* FPS calculated using time difference for 30 frames */
                }
                CyU3PDebugPrint(1,"\n\rFrm_Cnt = %d TimeDiff = %d ms FPS = %d", glFrameCount, glPeriod, fps);
            }

#endif
        }
#endif

        CyU3PThreadSleep(1);

        /* Wait for the events of interest */
        status = CyU3PEventGet(&glFx3Event, flagsOfInterest, CYU3P_EVENT_OR_CLEAR, &eventFlag, CYU3P_NO_WAIT);
        if (status == CY_U3P_SUCCESS)
        {
            switch(eventFlag)
            {

            case FX3_DMA_RESET_EVENT:
            {
                status = CyFx3UvcHandleDmaReset();
                if (status != CY_U3P_SUCCESS)
                {
                    CyU3PDebugPrint(4, "\n\rCyFx3UvcHandleDmaReset failed, Error code = 0x%x", status);
                }
#ifdef STILL_CAPTURE_ENABLE
                /*If the frame being flushed is Still image, the event is to be set to get back to video streaming;
                 * Check "Still Capture Trigger" part of code in CyFx3UvcAppUSBSetupCB function*/
                if (glStillFlag == CyTrue)
                {
                    CyU3PEventSet(&glStillImageEvent, FX3_STILL_IMAGE_EVENT, CYU3P_EVENT_OR);
                    CyU3PDebugPrint(4, "\n\rStill Capture Failed");
                }
#endif
            }
            break;

            case FX3_USB_SUSP_EVENT_FLAG:
            {
                /* Place FX3 in Low Power Suspend mode */
                CyU3PDebugPrint(4, "\n\rEntering LP Suspend Mode...");
                CyU3PThreadSleep(10);

                /* If the suspend event timer is cleared in the USB setupcb(), don't enter the suspend event */
                if(glReceivedSuspEvent)
                {
                    /* As per the USB3 specs, link layer takes 10ms (Max.) to leave U3 state. If the device sees some spurious (unwanted)
                     * USB activity it will leave suspend mode (even though USB is in U3 state). The device can wait for 10ms (U3 exit LFPS duration)
                     * and check the Link Layer State. If the link layer is not in U3 state, FX3 device will wake up else it will enter suspend mode */
                    do
                    {
                        status = CyU3PSysEnterSuspendMode(CY_U3P_SYS_USB_BUS_ACTVTY_WAKEUP_SRC, 0, &wakeReason);
                        if ((status != CY_U3P_SUCCESS) || (CyU3PUsbGetSpeed() != CY_U3P_SUPER_SPEED))
                            break;
                        /* Wait for the maximum U3 exit LFPS duration. */
                        CyU3PThreadSleep(10);

                        /* If the link is still in U3, we can continue to attempt Suspend mode entry. */
                        status = CyU3PUsbGetLinkPowerState(&usb3mode);
                        if ((status != CY_U3P_SUCCESS) || (usb3mode != CyU3PUsbLPM_U3))
                            break;
                    }while(1);
                }

                glReceivedSuspEvent = CyFalse;
                CyU3PDebugPrint(4, "\n\rLeaving Suspend Mode 0x%x, Wakeup reason = 0x%x", status, wakeReason);

                CyU3PThreadSleep(10);
            }
            break;

            default:
                break;
            }
        }
    }
}

/* Application define function which creates the threads. */
void CyFxApplicationDefine(void)
{
    void *ptr = NULL;
    uint32_t apiRetStatus = CY_U3P_SUCCESS;

    /* Allocate the memory for the thread and create the thread */
    ptr = CyU3PMemAlloc(UVC_APP_THREAD_STACK);
    if (ptr == NULL)
    {
        goto StartupError;
    }

    apiRetStatus = CyU3PThreadCreate(&uvcAppThread, /* UVC Thread structure */
            "30:UVC_app_thread",                    /* Thread Id and name */
            CyFx3UvcAppThread_Entry,                /* UVC Application Thread Entry function */
            0,                                      /* No input parameter to thread */
            ptr,                                    /* Pointer to the allocated thread stack */
            UVC_APP_THREAD_STACK,                   /* UVC Application Thread stack size */
            UVC_APP_THREAD_PRIORITY,                /* UVC Application Thread priority */
            UVC_APP_THREAD_PRIORITY,                /* Pre-emption threshold */
            CYU3P_NO_TIME_SLICE,                    /* No time slice for the application thread */
            CYU3P_AUTO_START                        /* Start the Thread immediately */
    );

    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        goto StartupError;
    }

    /* Create GPIO application event group */
    if (CyU3PEventCreate(&glFx3Event) != CY_U3P_SUCCESS)
    {
        goto StartupError;
    }

#ifdef STILL_CAPTURE_ENABLE
    /* Create stillcapture application event group for still image related events */
    if (CyU3PEventCreate(&glStillImageEvent) != CY_U3P_SUCCESS)
    {
        goto StartupError;
    }
#endif

    return;

    StartupError: {
        /* Failed to create threads and objects required for the application. This is a fatal error and we cannot
         * continue.
         */

        /* Add custom recovery or debug actions here */

        while (1);
    }
}

/*
 * Main function
 */
int main(void)
{
    CyU3PIoMatrixConfig_t io_cfg;
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    /* Initialize the device */
    CyU3PSysClockConfig_t clockConfig;
    clockConfig.setSysClk400 = CyTrue;
    clockConfig.cpuClkDiv = 2;
    clockConfig.dmaClkDiv = 2;
    clockConfig.mmioClkDiv = 2;
    clockConfig.useStandbyClk = CyFalse;
    clockConfig.clkSrc = CY_U3P_SYS_CLK;
    status = CyU3PDeviceInit(&clockConfig);
    if (status != CY_U3P_SUCCESS)
    {
        goto handle_fatal_error;
    }

    /* Initialize the caches. Enable instruction cache and keep data cache disabled.
     * The data cache is useful only when there is a large amount of CPU based memory
     * accesses. When used in simple cases, it can decrease performance due to large
     * number of cache flushes and cleans and also it adds to the complexity of the
     * code. */
    status = CyU3PDeviceCacheControl(CyTrue, CyFalse, CyFalse);
    if (status != CY_U3P_SUCCESS)
    {
        goto handle_fatal_error;
    }

    /* Configure the IO matrix for the device.*/
    io_cfg.isDQ32Bit = CyFalse;
    io_cfg.useUart   = CyTrue;
    io_cfg.useI2C    = CyTrue;
    io_cfg.useI2S    = CyFalse;
    io_cfg.useSpi    = CyTrue;

    io_cfg.lppMode = CY_U3P_IO_MATRIX_LPP_DEFAULT;

    /* No GPIOs are enabled. */
    io_cfg.gpioSimpleEn[0] = 0;
    io_cfg.gpioSimpleEn[1] = 0x02002000;		// GPIO 44 is ISP INT . This pin needs to be monitored to see if ISP is ready
    io_cfg.gpioComplexEn[0] = 0;
    io_cfg.gpioComplexEn[1] = 0;
    status = CyU3PDeviceConfigureIOMatrix(&io_cfg);
    if (status != CY_U3P_SUCCESS)
    {
        goto handle_fatal_error;
    }

    /* This is a non returnable call for initializing the RTOS kernel */
    CyU3PKernelEntry();

    /* Dummy return to make the compiler happy */
    return 0;

    handle_fatal_error:
    /* Cannot recover from this error. */
    while (1);
}

/* [ ] */

